from . import MailBaseSettings


class MailSettings(MailBaseSettings):
    mx_servers = [
        "1 aspmx.l.google.com.",
        "5 alt1.aspmx.l.google.com.",
        "5 alt2.aspmx.l.google.com.",
        "10 alt3.aspmx.l.google.com.",
        "10 alt4.aspmx.l.google.com.",
    ]


class Mail:
    ...
